"
 COPYRIGHT (c) 1997 by eXept Software AG
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

'From Smalltalk/X, Version:3.4.7 on 1-aug-1998 at 09:51:13'                     !

Filename subclass:#PCFilename
	instanceVariableNames:''
	classVariableNames:'StandardSuffixTable'
	poolDictionaries:''
	category:'OS-Windows'
!

!PCFilename class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1997 by eXept Software AG
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    Filenames in Windows-NT / Win95.
"
! !

!PCFilename class methodsFor:'initialization'!

initStandardSuffixTable
    "since there is no 'file' command to extract the type,
     return a guess based upon the files suffix. The following
     table defines what is returned."

    StandardSuffixTable := Dictionary new.
    #(  
	'COM'   'executable'
	'DIR'   'directory'
	'EXE'   'executable'
	'LST'   'listing'
	'OBJ'   'object file'
	'TMP'   'temporary'
	'BAS'   'basic source'
	'C'     'c source'
	'COB'   'cobol source'
	'FOR'   'fortran source'
	'PAS'   'pascal source'
	'PL1'   'PL/1 source'
	'ST'    'smalltalk source'
	'STH'   'stc generated header'
	'DLL'   'dynamic link library'
    ) pairWiseDo:[:k :v |
	StandardSuffixTable at:k put:v
    ]

    "
     StandardSuffixTable := nil.
     self initStandardSuffixTable
    "

    "Modified: 16.10.1997 / 13:12:39 / cg"
! !

!PCFilename class methodsFor:'instance creation'!

newTemporaryIn:aDirectoryPrefix
    "return a new unique filename - use this for temporary files.
     redefined to always return an MSDOS 8+3 fileName,
     in case the tempFile is passed to an OLD dos utility.."

    |pid nr nameString|

    (self ~~ ConcreteClass) ifTrue:[
	^ ConcreteClass newTemporaryIn:aDirectoryPrefix
    ].

    "/ although the above allows things to be redefined in concrete classes,
    "/ the following should work on all systems ...

    NextTempFilenameIndex isNil ifTrue:[
	NextTempFilenameIndex := 1.
    ].

    pid := OperatingSystem getProcessId printString.
    pid := pid copyLast:(3 min:pid size).
    nr := NextTempFilenameIndex printString.
    nr := nr copyLast:(3 min:nr size).
    nameString := (self tempFileNameTemplate)
		  bindWith:pid 
		  with:nr.
    NextTempFilenameIndex := NextTempFilenameIndex + 1.

    (aDirectoryPrefix isNil or:[aDirectoryPrefix asString isEmpty]) ifFalse:[
	^ aDirectoryPrefix asFilename construct:nameString
    ].
    ^ self named:nameString

    "temp files in '/tmp':

     Filename newTemporary    
    "

    "temp files somewhere 
     (not recommended - use above since it can be controlled via shell variables):

     Filename newTemporaryIn:'/tmp'    
     Filename newTemporaryIn:'/tmp'  
     Filename newTemporaryIn:'/usr/tmp'    
     Filename newTemporaryIn:'/'  
    "

    "a local temp file:

     Filename newTemporaryIn:''         
     Filename newTemporaryIn:nil         
     Filename newTemporaryIn:'.'         
     Filename newTemporaryIn:('source' asFilename) 
    "

    "Modified: 7.9.1995 / 10:48:31 / claus"
    "Modified: 8.9.1997 / 00:28:33 / cg"
    "Created: 30.1.1998 / 11:49:33 / md"
    "Modified: 30.1.1998 / 11:52:06 / md"
    "Modified: 30.1.1998 / 12:09:18 / dq"
! !

!PCFilename class methodsFor:'queries'!

isBadCharacter:aCharacter
    "return true, if aCharacter is unallowed in a filename."

    ('<>:"/\|' includes:aCharacter) ifTrue:[^ true].
    ^ super isBadCharacter:aCharacter

    "Created: 8.9.1997 / 00:14:06 / cg"
!

isCaseSensitive
    "return true, if filenames are case sensitive."

    ^ false
!

separator
    "return the file/directory separator."

     ^ $\

     "
      Filename concreteClass separator  
     "

    "Modified: 8.9.1997 / 00:18:03 / cg"
!

tempFileNameTemplate
    "return a template for temporary files.
     This is expanded with the current processID and a sequenceNumber
     to generate a unique filename.
     Redefined for MSDOS 8+3 filenames"

    ^ 'st%1%2.tmp'

    "Created: 30.1.1998 / 12:09:18 / dq"
! !

!PCFilename methodsFor:'file operations'!

renameTo:newName
    "rename the file - the argument must be convertable to a String.
     Raise an error if not successful.
     Redefined to delete any existing target-file first."

    newName asFilename exists ifTrue:[
	newName asFilename delete
    ].

    ^ super renameTo:newName

    "
     '\tmp\foo' asFilename renameTo:'\tmp\bar'
    "

    "Modified: 20.1.1998 / 15:33:00 / md"
! !

!PCFilename methodsFor:'queries'!

fileType
    "this returns a string describing the type of contents of
     the file. Here, the suffix is examined for a standard
     suffix and an appropriate string is returned.
     Poor MSDOS - no file command."

    |suff type info fmt|

    StandardSuffixTable isNil ifTrue:[
	self class initStandardSuffixTable
    ].

    suff := self suffix asUppercase.
    type := StandardSuffixTable at:suff ifAbsent:nil.
    type isNil ifTrue:[
	type := super fileType.
    ].
    ^ type

    "Created: 16.10.1997 / 13:07:24 / cg"
    "Modified: 16.10.1997 / 13:10:00 / cg"
!

isExecutableProgram
    "return true, if such a file exists and is an executable program.
     (i.e. for directories, false is returned.)"

    ((self hasSuffix:'exe') 
    or:[self hasSuffix:'com']) ifTrue:[
	^ super isExecutableProgram
    ].
    ^ false

    "Created: 16.10.1997 / 13:19:10 / cg"
!

isExplicitRelative
    "return true, if this name is an explicit relative name
     (i.e. starts with './' or '../', to avoid path-prepending)"

    (nameString startsWith:'.\') ifTrue:[
	^ true
    ].
    (nameString startsWith:'..\') ifTrue:[
	^ true
    ].
    ^ false
!

isVolumeAbsolute
    "return true, if the receiver represents an absolute pathname
     on some disk volume."

    ^ (((nameString size >= 3) and:[(nameString at:2) == $:]) and:[(nameString at:3) == $\])

    "Created: 7.9.1997 / 23:54:20 / cg"
!

localNameStringFrom:aString 
    "ST-80 compatibility.
     what does this do ? (used in FileNavigator-goody).
     GUESS: does it strip off the voulume-character and initial '\' ?"

    (aString at:2) == $: ifTrue:[
	(aString at:3) == $\ ifTrue:[
	    ^ aString copyFrom:4
	].
	^ aString copyFrom:3
    ]. 
    (aString at:1) == $\ ifTrue:[
	^ aString copyFrom:1
    ].
    ^ aString
!

volume
    "return the disc volume part of the name or an empty string.
     This is only used with DOS filenames - on unix, an empty string is returned"

    (nameString at:2) == $: ifTrue:[
	^ nameString copyTo:1
    ].
    ^ ''

    "Created: 7.9.1997 / 23:58:06 / cg"
! !

!PCFilename methodsFor:'special accessing'!

osNameForDirectory
    "special - return the OS's name for the receiver to
     access it as a directory.
     Care remove trailing backSlashes here"

    (nameString endsWith:'\') ifTrue:[
        ((nameString size == 3) and:[(nameString at:2) == $:]) ifFalse:[
            ^nameString copyWithoutLast:1
        ]
    ].
    ^ nameString

    "Modified: 20.1.1998 / 15:39:06 / md"
!

setName:aString
    "set the filename"

    nameString := aString copy replaceAll:$/ with:$\

    "Created: 22.1.1998 / 17:32:45 / md"
! !

!PCFilename class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libbasic/PCFilename.st,v 1.9 1998-06-15 13:39:10 cg Exp $'
! !
